"""API Helper."""
import json
from datetime import datetime
from typing import Any

from requests import Response, session

from .const import (
    SUNWEG_INVERTER_DETAIL_PATH,
    SUNWEG_LOGIN_PATH,
    SUNWEG_PLANT_DETAIL_PATH,
    SUNWEG_PLANT_LIST_PATH,
    SUNWEG_URL,
)
from .device import MPPT, Inverter, Phase, String
from .plant import Plant
from .util import SingletonMeta, Status


class SunWegApiError(RuntimeError):
    """API Error."""

    pass


class LoginError(SunWegApiError):
    """Login Error."""

    pass


class APIHelper(metaclass=SingletonMeta):
    """Singleton class to call sunweg.net api."""

    SERVER_URI = SUNWEG_URL

    def __init__(self, username: str, password: str) -> None:
        """
        Initialize APIHelper for SunWEG platform.

        :param username: username for authentication
        :param password: password for authentication
        :type username: str
        :type password: str
        """
        self._token = None
        self._username = username
        self._password = password
        self.session = session()

    def authenticate(self) -> bool:
        """
        Authenticate with provided username and password.

        :return: True on authentication success
        :rtype: bool
        """
        user_data = json.dumps(
            {"usuario": self._username, "senha": self._password},
            default=lambda o: o.__dict__,
        )

        result = self._post(SUNWEG_LOGIN_PATH, user_data)
        self._token = result["token"]
        return result["success"]

    def _headers(self):
        """Retrieve headers with authentication token."""
        if self._token is None:
            return {}
        return {"X-Auth-Token-Update": self._token}

    def listPlants(self, retry=True) -> list[Plant]:
        """
        Retrieve the list of plants with incomplete inverter information.

        You may want to call `complete_inverter()` to complete the Inverter information.

        :param retry: reauthenticate if token expired and retry
        :type retry: bool
        :return: list of Plant
        :rtype: list[Plant]
        """
        try:
            result = self._get(SUNWEG_PLANT_LIST_PATH)
            ret_list = []
            for plant in result["usinas"]:
                if (plant := self.plant(plant["id"])) is not None:
                    ret_list.append(plant)
            return ret_list
        except LoginError:
            if retry:
                self.authenticate()
                return self.listPlants(False)
            return []

    def plant(self, id: int, retry=True) -> Plant | None:
        """
        Retrieve plant detail by plant id.

        :param id: plant id
        :type id: int
        :param retry: reauthenticate if token expired and retry
        :type retry: bool
        :return: Plant or None if `id` not found.
        :rtype: Plant | None
        """
        try:
            result = self._get(SUNWEG_PLANT_DETAIL_PATH + str(id))

            plant = Plant(
                id=id,
                name=result["usinas"]["nome"],
                total_power=float(
                    str(result["AcumuladoPotencia"])
                    .replace(" kW", "")
                    .replace(",", ".")
                ),
                kwh_per_kwp=float(str(result["KWHporkWp"]).replace(",", "."))
                if result["KWHporkWp"] != ""
                else float(0),
                performance_rate=result["taxaPerformance"],
                saving=result["economia"],
                today_energy=float(
                    str(result["energiaGeradaHoje"])
                    .replace(" kWh", "")
                    .replace(",", ".")
                ),
                total_energy=float(result["energiaacumuladanumber"]),
                total_carbon_saving=result["reduz_carbono_total_number"],
                last_update=datetime.strptime(
                    result["ultimaAtualizacao"], "%Y-%m-%d %H:%M:%S"
                ),
            )

            plant.inverters.extend(
                [
                    Inverter(
                        id=inv["id"],
                        name=inv["nome"],
                        sn=inv["esn"],
                        status=Status(int(inv["situacao"])),
                        temperature=inv["temperatura"],
                    )
                    for inv in result["usinas"]["inversores"]
                ]
            )
            return plant
        except LoginError:
            if retry:
                self.authenticate()
                return self.plant(id, False)
            return None

    def inverter(self, id: int, retry=True) -> Inverter | None:
        """
        Retrieve inverter detail by inverter id.

        :param id: inverter id
        :type id: int
        :param retry: reauthenticate if token expired and retry
        :type retry: bool
        :return: Inverter or None if `id` not found.
        :rtype: Inverter | None
        """
        try:
            result = self._get(SUNWEG_INVERTER_DETAIL_PATH + str(id))
            inverter = Inverter(
                id=id,
                name=result["inversor"]["nome"],
                sn=result["inversor"]["esn"],
                total_energy=float(
                    result["energiaAcumulada"].replace(" kWh", "").replace(",", ".")
                ),
                today_energy=float(
                    result["energiaDoDia"].replace(" kWh", "").replace(",", ".")
                ),
                power_factor=float(result["fatorpotencia"].replace(",", ".")),
                frequency=result["frequencia"],
                power=float(
                    result["potenciaativa"].replace(" kW", "").replace(",", ".")
                ),
                status=Status(int(result["statusInversor"])),
                temperature=result["temperatura"],
            )

            self._populate_MPPT(result=result, inverter=inverter)

            return inverter
        except LoginError:
            if retry:
                self.authenticate()
                return self.inverter(id, False)
            return None

    def complete_inverter(self, inverter: Inverter, retry=True) -> None:
        """
        Complete inverter data.

        :param inverter: inverter object to be completed with information
        :type inverter: Inverter
        :param retry: reauthenticate if token expired and retry
        :type retry: bool
        """
        try:
            result = self._get(SUNWEG_INVERTER_DETAIL_PATH + str(inverter.id))
            inverter.total_energy = float(
                result["energiaAcumulada"].replace(" kWh", "").replace(",", ".")
            )
            inverter.today_energy = float(
                result["energiaDoDia"].replace(" kWh", "").replace(",", ".")
            )
            inverter.power_factor = float(result["fatorpotencia"].replace(",", "."))
            inverter.frequency = result["frequencia"]
            inverter.power = float(
                result["potenciaativa"].replace(" kW", "").replace(",", ".")
            )

            self._populate_MPPT(result=result, inverter=inverter)
        except LoginError:
            if retry:
                self.authenticate()
                self.complete_inverter(inverter, False)

    def _populate_MPPT(self, result: dict, inverter: Inverter) -> None:
        """Populate MPPT information inside a inverter."""
        for str_mppt in result["stringmppt"]:
            mppt = MPPT(str_mppt["nomemppt"])

            for str_string in str_mppt["strings"]:
                string = String(
                    str_string["nome"],
                    float(str_string["valorTensao"]),
                    float(str_string["valorCorrente"]),
                    Status(int(str_string["status"])),
                )
                mppt.strings.append(string)

            inverter.mppts.append(mppt)

        for phase_name in result["correnteCA"].keys():
            if str(phase_name).endswith("status"):
                continue
            inverter.phases.append(
                Phase(
                    phase_name,
                    float(result["tensaoca"][phase_name].replace(",", ".")),
                    float(result["correnteCA"][phase_name].replace(",", ".")),
                    Status(result["tensaoca"][phase_name + "status"]),
                    Status(result["correnteCA"][phase_name + "status"]),
                )
            )

    def _get(self, path: str) -> dict:
        """Do a get request returning a treated response."""
        res = self.session.get(self.SERVER_URI + path, headers=self._headers())
        result = self._treat_response(res)
        return result

    def _post(self, path: str, data: Any | None) -> dict:
        """Do a post request returning a treated response."""
        res = self.session.post(
            self.SERVER_URI + path, data=data, headers=self._headers()
        )
        result = self._treat_response(res)
        return result

    def _treat_response(self, response: Response) -> dict:
        """Treat the response from requests."""
        if response.status_code == 401:
            raise LoginError("Request failed: %s" % response)
        if response.status_code != 200:
            raise SunWegApiError("Request failed: %s" % response)
        result = response.json()
        if not result["success"]:
            raise SunWegApiError(result["message"])
        return result



"""Test sunweg common."""
from datetime import datetime
from sunweg.device import Inverter, MPPT, Phase, String
from sunweg.plant import Plant
from sunweg.util import Status


PLANT_MOCK = Plant(
    id=1,
    name="Plant",
    total_power=29.2,
    kwh_per_kwp=0.1,
    performance_rate=0,
    saving=0,
    today_energy=123.1,
    total_energy=321.1,
    total_carbon_saving=12.1,
    last_update=datetime.now(),
)

INVERTER_MOCK = Inverter(
    id=1,
    name="Inverter",
    sn="1234ABC",
    total_energy=321.1,
    today_energy=123.1,
    power_factor=0.2,
    frequency=60,
    power=29,
    status=Status.OK,
    temperature=70,
)

MPPT_MOCK = MPPT("MPPT")

STRING_MOCK = String(name="String", voltage=523.1, amperage=12.1, status=Status.OK)

PHASE_MOCK = Phase(
    name="Phase",
    voltage=230.1,
    amperage=3.1,
    status_voltage=Status.OK,
    status_amperage=Status.OK,
)



"""Test sunweg.api."""
from datetime import datetime
from os import path
import os
from unittest import TestCase
from unittest.mock import patch
import pytest

from requests import Response

from sunweg.api import APIHelper, SunWegApiError
from sunweg.device import Inverter, String
from sunweg.util import Status

from .common import INVERTER_MOCK, PLANT_MOCK


class Api_Test(TestCase):
    """APIHelper test case."""

    responses: dict[str, Response] = {}

    def setUp(self) -> None:
        """Set tests up."""
        for file in os.listdir(path.join(path.dirname(__file__), "responses")):
            filename = path.basename(file)
            with open(path.join(path.dirname(__file__), "responses", file)) as f:
                response = Response()
                if filename.startswith("error"):
                    response.status_code = int(filename.split("_")[1])
                    response.reason = "".join(f.readlines())
                else:
                    response.status_code = 200
                    response._content = "".join(f.readlines()).encode()
                self.responses[filename] = response

    def test_error500(self) -> None:
        """Test error 500."""
        with patch(
            "requests.Session.post",
            return_value=self.responses["error_500_response.txt"],
        ):
            api = APIHelper("user@acme.com", "password")
            with pytest.raises(SunWegApiError) as e_info:
                api.authenticate()
            assert e_info.value.__str__() == "Request failed: <Response [500]>"

    def test_authenticate_success(self) -> None:
        """Test authentication success."""
        with patch(
            "requests.Session.post",
            return_value=self.responses["auth_success_response.json"],
        ):
            api = APIHelper("user@acme.com", "password")
            assert api.authenticate()

    def test_authenticate_failed(self) -> None:
        """Test authentication failed."""
        with patch(
            "requests.Session.post",
            return_value=self.responses["auth_fail_response.json"],
        ):
            api = APIHelper("user@acme.com", "password")
            with pytest.raises(SunWegApiError) as e_info:
                api.authenticate()
            assert e_info.value.__str__() == "Error message"

    def test_list_plants_none_success(self) -> None:
        """Test list plants with empty plant list."""
        with patch(
            "requests.Session.get",
            return_value=self.responses["list_plant_success_none_response.json"],
        ), patch("sunweg.api.APIHelper.plant", return_value=PLANT_MOCK):
            api = APIHelper("user@acme.com", "password")
            assert len(api.listPlants()) == 0

    def test_list_plants_1_success(self) -> None:
        """Test list plants with one plant in the list."""
        with patch(
            "requests.Session.get",
            return_value=self.responses["list_plant_success_1_response.json"],
        ), patch("sunweg.api.APIHelper.plant", return_value=PLANT_MOCK):
            api = APIHelper("user@acme.com", "password")
            assert len(api.listPlants()) == 1

    def test_list_plants_2_success(self) -> None:
        """Test list plants with two plant in the list."""
        with patch(
            "requests.Session.get",
            return_value=self.responses["list_plant_success_2_response.json"],
        ), patch("sunweg.api.APIHelper.plant", return_value=PLANT_MOCK):
            api = APIHelper("user@acme.com", "password")
            assert len(api.listPlants()) == 2

    def test_list_plants_401(self) -> None:
        """Test list plants with expired token."""
        with patch(
            "requests.Session.post",
            return_value=self.responses["auth_success_response.json"],
        ), patch(
            "requests.Session.get",
            return_value=self.responses["error_401_response.txt"],
        ):
            api = APIHelper("user@acme.com", "password")
            assert len(api.listPlants()) == 0

    def test_plant_success(self) -> None:
        """Test plant success."""
        with patch(
            "requests.Session.get",
            return_value=self.responses["plant_success_response.json"],
        ), patch("sunweg.api.APIHelper.inverter", return_value=INVERTER_MOCK):
            api = APIHelper("user@acme.com", "password")
            plant = api.plant(16925)
            assert plant is not None
            assert plant.id == 16925
            assert plant.name == "Plant Name"
            assert plant.total_power == 25.23
            assert plant.last_update == datetime(2023, 2, 25, 8, 4, 22)
            assert plant.kwh_per_kwp == 1.2
            assert plant.performance_rate == 1.48
            assert plant.saving == 12.786912
            assert plant.today_energy == 1.23
            assert plant.total_carbon_saving == 0.012296
            assert plant.total_energy == 23.2
            assert plant.__str__().startswith("<class 'sunweg.plant.Plant'>")
            assert len(plant.inverters) == 1
            for inverter in plant.inverters:
                assert inverter.id == 21255
                assert inverter.name == "Inverter Name"
                assert inverter.frequency == 0
                assert inverter.power == 0.0
                assert inverter.power_factor == 0.0
                assert inverter.sn == "1234ABC"
                assert inverter.status == Status.ERROR
                assert inverter.temperature == 80
                assert inverter.today_energy == 0.0
                assert inverter.total_energy == 0.0
                assert not inverter.is_complete

    def test_plant_401(self) -> None:
        """Test plant with expired token."""
        with patch(
            "requests.Session.post",
            return_value=self.responses["auth_success_response.json"],
        ), patch(
            "requests.Session.get",
            return_value=self.responses["error_401_response.txt"],
        ):
            api = APIHelper("user@acme.com", "password")
            assert api.plant(16925) is None

    def test_inverter_success(self) -> None:
        """Test inverter success."""
        with patch(
            "requests.Session.get",
            return_value=self.responses["inverter_success_response.json"],
        ):
            api = APIHelper("user@acme.com", "password")
            inverter = api.inverter(21255)
            assert inverter is not None
            assert inverter.id == 21255
            assert inverter.name == "Inverter Name"
            assert inverter.frequency == 0
            assert inverter.power == 0.0
            assert inverter.power_factor == 0.0
            assert inverter.sn == "1234ABC"
            assert inverter.status == Status.OK
            assert inverter.temperature == 80
            assert inverter.today_energy == 0.0
            assert inverter.total_energy == 23.2
            strings: list[String] = []
            for mppt in inverter.mppts:
                assert mppt.__str__().startswith("<class 'sunweg.device.MPPT'>")
                assert mppt.name != ""
                strings.extend(mppt.strings)
            assert len(strings) == 4
            assert len(inverter.phases) == 3
            assert inverter.__str__().startswith("<class 'sunweg.device.Inverter'>")
            for string in strings:
                assert string.name != ""
                assert string.amperage != 0
                assert string.voltage != 0
                assert string.status == Status.WARN
                assert string.__str__().startswith("<class 'sunweg.device.String'>")
            for phase in inverter.phases:
                assert phase.name != ""
                assert phase.amperage != 0
                assert phase.voltage != 0
                assert phase.status_amperage == Status.OK
                assert phase.status_voltage == Status.ERROR
                assert phase.__str__().startswith("<class 'sunweg.device.Phase'>")

    def test_inverter_401(self) -> None:
        """Test inverter with expired token."""
        with patch(
            "requests.Session.post",
            return_value=self.responses["auth_success_response.json"],
        ), patch(
            "requests.Session.get",
            return_value=self.responses["error_401_response.txt"],
        ):
            api = APIHelper("user@acme.com", "password")
            assert api.inverter(21255) is None

    def test_complete_inverter_success(self) -> None:
        """Test complete inverter success."""
        with patch(
            "requests.Session.get",
            return_value=self.responses["inverter_success_response.json"],
        ):
            api = APIHelper("user@acme.com", "password")
            inverter = Inverter(
                id=12345,
                name="Other inverter name",
                sn="1234ABCD",
                status=Status.ERROR,
                temperature=70,
            )
            api.complete_inverter(inverter)
            assert inverter is not None
            assert inverter.id == 12345
            assert inverter.name == "Other inverter name"
            assert inverter.frequency == 0
            assert inverter.power == 0.0
            assert inverter.power_factor == 0.0
            assert inverter.sn == "1234ABCD"
            assert inverter.status == Status.ERROR
            assert inverter.temperature == 70
            assert inverter.today_energy == 0.0
            assert inverter.total_energy == 23.2
            strings: list[String] = []
            for mppt in inverter.mppts:
                assert mppt.__str__().startswith("<class 'sunweg.device.MPPT'>")
                assert mppt.name != ""
                strings.extend(mppt.strings)
            assert len(strings) == 4
            assert len(inverter.phases) == 3
            assert inverter.__str__().startswith("<class 'sunweg.device.Inverter'>")
            for string in strings:
                assert string.name != ""
                assert string.amperage != 0
                assert string.voltage != 0
                assert string.status == Status.WARN
                assert string.__str__().startswith("<class 'sunweg.device.String'>")
            for phase in inverter.phases:
                assert phase.name != ""
                assert phase.amperage != 0
                assert phase.voltage != 0
                assert phase.status_amperage == Status.OK
                assert phase.status_voltage == Status.ERROR
                assert phase.__str__().startswith("<class 'sunweg.device.Phase'>")

    def test_complete_inverter_401(self) -> None:
        """Test complete inverter with expired token."""
        with patch(
            "requests.Session.post",
            return_value=self.responses["auth_success_response.json"],
        ), patch(
            "requests.Session.get",
            return_value=self.responses["error_401_response.txt"],
        ):
            api = APIHelper("user@acme.com", "password")
            inverter = Inverter(
                id=12345,
                name="Other inverter name",
                sn="1234ABCD",
                status=Status.ERROR,
                temperature=70,
            )
            api.complete_inverter(inverter)
            assert not inverter.is_complete



"""Sunweg API devices."""
from .util import Status


class Phase:
    """Phase details."""

    def __init__(
        self,
        name: str,
        voltage: float,
        amperage: float,
        status_voltage: Status,
        status_amperage: Status,
    ) -> None:
        """
        Initialize Phase.

        :param name: phase name
        :type name: str
        :param voltage: phase AC voltage in V
        :type voltage: float
        :param amperage: phase AC amperage in A
        :type amperage: float
        :param status_voltage: phase AC voltage status
        :type status_voltage: Status
        :param status_amperage: phase AC amperage status
        :type status_amperage: Status
        """
        self._name = name
        self._voltage = voltage
        self._amperage = amperage
        self._status_voltage = status_voltage
        self._status_amperage = status_amperage

    @property
    def name(self) -> str:
        """
        Get phase name.

        :return: phase name
        :rtype: str
        """
        return self._name

    @property
    def voltage(self) -> float:
        """
        Get phase AC voltage in V.

        :return: phase AC voltage in V
        :rtype: float
        """
        return self._voltage

    @property
    def amperage(self) -> float:
        """
        Get phase AC amperage in A.

        :return: phase AC amperage in A
        :rtype: float
        """
        return self._amperage

    @property
    def status_voltage(self) -> Status:
        """
        Get phase AC voltage status.

        :return: phase AC voltage status
        :rtype: Status
        """
        return self._status_voltage

    @property
    def status_amperage(self) -> Status:
        """
        Get phase AC amperage status.

        :return: phase AC amperage status
        :rtype: Status
        """
        return self._status_amperage

    def __str__(self) -> str:
        """Cast Phase to str."""
        return str(self.__class__) + ": " + str(self.__dict__)


class String:
    """String details."""

    def __init__(
        self, name: str, voltage: float, amperage: float, status: Status
    ) -> None:
        """
        Initialize String.

        :param name: string name
        :type name: str
        :param voltage: string DC voltage in V
        :type voltage: float
        :param amperage: string DC amperage in A
        :type amperage: float
        :param status: string status
        :type status: Status
        """
        self._name = name
        self._voltage = voltage
        self._amperage = amperage
        self._status = status

    @property
    def name(self) -> str:
        """
        Get string name.

        :return: string name
        :rtype: str
        """
        return self._name

    @property
    def voltage(self) -> float:
        """
        Get string DC voltage in V.

        :return: string DC voltage in V
        :rtype: float
        """
        return self._voltage

    @property
    def amperage(self) -> float:
        """
        Get string DC amperage in A.

        :return: string DC amperage in A
        :rtype: float
        """
        return self._amperage

    @property
    def status(self) -> Status:
        """
        Get string status.

        :return: string status
        :rtype: Status
        """
        return self._status

    def __str__(self) -> str:
        """Cast String to str."""
        return str(self.__class__) + ": " + str(self.__dict__)


class MPPT:
    """MPPT details."""

    def __init__(self, name: str) -> None:
        """
        Initialize MPPT.

        :param name: MPPT name
        :type name: srt
        """
        self._name = name
        self._strings: list[String] = []

    @property
    def name(self) -> str:
        """
        Get MPPT name.

        :return: MPPT name
        :rtype: str
        """
        return self._name

    @property
    def strings(self) -> list[String]:
        """
        Get list of MPPT's String.

        :return: list of Strings
        :rtype: list[String]
        """
        return self._strings

    def __str__(self) -> str:
        """Cast MPPT to str."""
        return str(self.__class__) + ": " + str(self.__dict__)


class Inverter:
    """Inverter device."""

    def __init__(
        self,
        id: int,
        name: str,
        sn: str,
        status: Status,
        temperature: int,
        total_energy: float = 0,
        today_energy: float = 0,
        power_factor: float = 0,
        frequency: float = 0,
        power: float = 0,
    ) -> None:
        """
        Initialize Inverter.

        :param id: inverter id
        :type id: int
        :param name: inverter name
        :type name: str
        :param sn: inverter serial number
        :type sn: str
        :param status: inverter status
        :type status: Status
        :param temperature: inverter temperature
        :type temperature: int
        :param total_energy: total generated energy in kWh
        :type total_energy: float
        :param today_energy: total generated energy today in kWh
        :type today_energy: float
        :param power_factor: inverter power factor
        :type power_factor: float
        :param frequency: inverter output frequency in Hz
        :type frequency: float
        :param power: inverter output power in W
        :type power: float
        """
        self._id = id
        self._name = name
        self._sn = sn
        self._total_energy = total_energy
        self._today_energy = today_energy
        self._power_factor = power_factor
        self._frequency = frequency
        self._power = power
        self._status = status
        self._temperature = temperature
        self._phases: list[Phase] = []
        self._mppts: list[MPPT] = []

    @property
    def id(self) -> int:
        """
        Get inverter id.

        :return: inverter id
        :rtype: int
        """
        return self._id

    @property
    def name(self) -> str:
        """
        Get inverter name.

        :return: inverter name
        :rtype: str
        """
        return self._name

    @property
    def sn(self) -> str:
        """
        Get inverter serial number.

        :return: inverter serial number
        :rtype: str
        """
        return self._sn

    @property
    def status(self) -> Status:
        """
        Get inverter status.

        :return: inverter status
        :rtype: Status
        """
        return self._status

    @property
    def temperature(self) -> int:
        """
        Get inverter temperature.

        :return: inverter temperature
        :rtype: int
        """
        return self._temperature

    @property
    def today_energy(self) -> float:
        """
        Get inverter today generated energy in kWh.

        :return: inverter today generated energy in kWh
        :rtype: float
        """
        return self._today_energy

    @today_energy.setter
    def today_energy(self, value: float) -> None:
        """
        Set inverter today generated energy in kWh.

        :param value: inverter today generated energy in kWh
        :type value: float
        """
        self._today_energy = value

    @property
    def total_energy(self) -> float:
        """
        Get inverter total generated energy in kWh.

        :return: inverter total generated energy in kWh
        :rtype: float
        """
        return self._total_energy

    @total_energy.setter
    def total_energy(self, value: float) -> None:
        """
        Set inverter total generated energy in kWh.

        :param value: inverter total generated energy in kWh
        :type value: float
        """
        self._total_energy = value

    @property
    def power_factor(self) -> float:
        """
        Get inverter power factor.

        :return: inverter power factor
        :rtype: float
        """
        return self._power_factor

    @power_factor.setter
    def power_factor(self, value: float) -> None:
        """
        Set inverter power factor.

        :param value: inverter power factor
        :type value: float
        """
        self._power_factor = value

    @property
    def frequency(self) -> float:
        """
        Get inverter frequency in Hz.

        :return: inverter frequency in HZ
        :rtype: float
        """
        return self._frequency

    @frequency.setter
    def frequency(self, value: float) -> None:
        """
        Set inverter frequency in Hz.

        :param value: inverter frequency in Hz
        :type value: float
        """
        self._frequency = value

    @property
    def power(self) -> float:
        """
        Get inverter output power in W.

        :return: inverter output power in W
        :rtype: float
        """
        return self._power

    @power.setter
    def power(self, value: float) -> None:
        """
        Set inverter output power in W.

        :param value: inverter output power in W
        :type value: float
        """
        self._power = value

    @property
    def is_complete(self) -> bool:
        """
        Is inverter data complete.

        :return: True when inverter data is complete
        :rtype: bool
        """
        return (
            self._today_energy != 0
            or self._total_energy != 0
            or self._power_factor != 0
            or self._frequency != 0
            or self._power != 0
        )

    @property
    def phases(self) -> list[Phase]:
        """
        Get list of inverter's phases.

        :return: list of phases
        :rtype: list[Phase]
        """
        return self._phases

    @property
    def mppts(self) -> list[MPPT]:
        """
        Get list of inverter's MPPTs.

        :return: list of MPPTs
        :rtype: list[MPPT]
        """
        return self._mppts

    def __str__(self) -> str:
        """Cast Inverter to str."""
        return str(self.__class__) + ": " + str(self.__dict__)



"""Sunweg API constants."""
SUNWEG_URL = "https://api.sun.weg.net/"
"""SunWEG API URL"""
SUNWEG_LOGIN_PATH = "login/autenticacao"
"""SunWEG API login path"""
SUNWEG_PLANT_LIST_PATH = (
    "usinas/getpaineloperacao?procurar="
    + "&integrador=&franqueado=&portal=&alarme=&"
    + "planos=%5B%220%22,%221%22,%222%22,%223%22,%224%22%5D&"
    + "status=%5B%221%22,%222%22,%223%22,%224%22,%225%22%5D"
)
"""SunWEG API list plants path"""
SUNWEG_PLANT_DETAIL_PATH = "usinas/view?idusina="
"""SunWEG API plant details path"""
SUNWEG_INVERTER_DETAIL_PATH = "inversores/view?id="
"""SunWEG API inverter details path"""



"""Sunweg API plant."""
from datetime import datetime

from .device import Inverter


class Plant:
    """Plant details."""

    def __init__(
        self,
        id: int,
        name: str,
        total_power: float,
        kwh_per_kwp: float,
        performance_rate: float,
        saving: float,
        today_energy: float,
        total_energy: float,
        total_carbon_saving: float,
        last_update: datetime,
    ) -> None:
        """
        Initialize Plant.

        :param id: plant id
        :type id: int
        :param name: plant name
        :type name: str
        :param total_power: plant total power
        :type total_power: float
        :param kwh_per_kwp: plant kWh/kWp
        :type kwh_per_kwp: float
        :param performance_rate: plant performance rate
        :type performance_rate: float
        :param saving: total saving in R$
        :type saving: float
        :param today_energy: today generated energy in kWh
        :type today_energy: float
        :param total_energy: total generated energy in kWh
        :type total_energy: float
        :param total_carbon_saving: total of CO2 saved
        :type total_carbon_saving: float
        :param last_update: when the data was updated
        :type last_update: datetime
        """
        self._id = id
        self._name = name
        self._total_power = total_power
        self._kwh_per_kwp = kwh_per_kwp
        self._performance_rate = performance_rate
        self._saving = saving
        self._today_energy = today_energy
        self._total_energy = total_energy
        self._total_carbon_saving = total_carbon_saving
        self._last_update = last_update
        self._inverters: list[Inverter] = []

    @property
    def id(self) -> int:
        """
        Get plant id.

        :return: plant id
        :rtype: int
        """
        return self._id

    @property
    def name(self) -> str:
        """
        Get plant name.

        :return: plant name
        :rtype: str
        """
        return self._name

    @property
    def total_power(self) -> float:
        """
        Get plant total power.

        :return: plant total power
        :rtype: float
        """
        return self._total_power

    @property
    def kwh_per_kwp(self) -> float:
        """
        Get plant kWh/kWp.

        :return: plant kWh/kWp
        :rtype: float
        """
        return self._kwh_per_kwp

    @property
    def performance_rate(self) -> float:
        """
        Get plant performance rate.

        :return: plant performance rate
        :rtype: float
        """
        return self._performance_rate

    @property
    def saving(self) -> float:
        """
        Get plant saving in R$.

        :return: plant saving in R$
        :rtype: float
        """
        return self._saving

    @property
    def today_energy(self) -> float:
        """
        Get plant today generated energy in kWh.

        :return: plant today generated energy in kWh
        :rtype: float
        """
        return self._today_energy

    @property
    def total_energy(self) -> float:
        """
        Get plant total generated energy in kWh.

        :return: plant total generated energy in kWh
        :rtype: float
        """
        return self._total_energy

    @property
    def total_carbon_saving(self) -> float:
        """
        Get plant total of CO2 saved.

        :return: plant total of CO2 saved
        :rtype: float
        """
        return self._total_carbon_saving

    @property
    def last_update(self) -> datetime:
        """
        Get when the plant data was updated.

        :return: when the plant data was updated
        :rtype: datetime
        """
        return self._last_update

    @property
    def inverters(self) -> list[Inverter]:
        """
        Get list of plant's inverters.

        :return: list of inverters
        :rtype: list[Inverter]
        """
        return self._inverters

    def __str__(self) -> str:
        """Cast Plant to str."""
        return str(self.__class__) + ": " + str(self.__dict__)



"""Sunweg API util."""
from enum import Enum


class Status(Enum):
    """Status enum."""

    OK = 0
    WARN = 2
    ERROR = 1


class SingletonMeta(type):
    """Singleton meta."""

    _instances: dict = {}

    def __call__(cls, *args, **kwargs):
        """Handle singleton creation."""
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]
